#ifndef _T_DB_RESULT_HH
#define _T_DB_RESULT_HH

#include <libpq-fe.h>
#include "TString.hh"

/**
 * Holds the result of a database query or some other database
 * related operation.
 *
 * Currently only PostgreSQL is supported.
 *
 * @author Johan Bolmsjo <johan@nocrew.org>
 */

class TDBResult
{
public:
  
  /**
   * Destructor.
   */
  ~TDBResult ();
  
  /**
   * Query if an error has occurred.
   *
   * @return  true, false
   */
  inline bool IsValid () { return myResult; }
  
  /**
   * Get error message string.
   *
   * @return  Error message.
   */
  TString& ErrorMessage ();
  
  /**
   * Query number of tuples in result.
   *
   * @return  Number of tuples.
   */
  inline int Tuples () { return myTuples; }
  
  /**
   * Query number of fields in result.
   *
   * @return  Number of fields.
   */
  inline int Fields () { return myFields; }
  
  /**
   * Query if specified field is null.
   *
   * @return  true, false
   */
  inline bool IsNull (int tuple, int field)
  {
    return PQgetisnull (myResult, tuple, field);
  }
  
  /**
   * Get value of the specified field.
   *
   * @param tuple  Tuple (row) number
   * @param field  Field (attribute) number
   * @return       Result string
   */
  TString& Value (int tuple, int field);
  
  /**
   * Get value of the specified field.
   * Unlike Value() the caller owns the returned string and is
   * responsible for destroying it when no longer needed.
   *
   * @param tuple  Tuple (row) number
   * @param field  Field (attribute) number
   * @return       Result string
   */
  TString* Value2 (int tuple, int field);
  
private:
  
  friend class TDBConnection;
  
  /* FIXME: No copying for now! */
  TDBResult (const TDBResult&);
  TDBResult& operator=(const TDBResult&);
  
  /**
   * Constructs object.
   * Initialises myTuples, myFields, myResult.
   * Call IsValid() to check error condition.
   */
  TDBResult (PGresult* result);
  
  /* Error message. */
  TString *myErrorMessage;
  
  /* Result from Value() */
  TString *myValue;
  
  /* PostgreSQL result */
  PGresult* myResult;
  
  /* Number of tuples in result. */
  int myTuples;
  
  /* Number of fields in result. */
  int myFields;
};

#endif // _T_DB_RESULT_HH
