#ifndef _T_FILE_HH
#define _T_FILE_HH

#include "TString.hh"
#if defined (_WIN32) || defined (__CYGWIN__)
#include <windows.h>
#endif

#define TFILE_BUF_LEN  4096


/**
 * File access class.
 * FIXME: Add copy operator and reference counting.
 */

class TFile
{
public:
  
  /**
   * Starting point used when seeking in file.
   */
  enum Seek
  {
    /** Seek from begining. */
    SET,
    
    /** Seek from current position. */
    CUR,
    
    /** Seek from end. */
    END 
  };
  
  /**
   * Default constructor.
   */
  TFile ();
  
  /**
   * Open the specified file.
   * Use IsValid() to check for validity.
   *
   * @param path  Path to file.
   * @param mode  Mode can be one of the following,
   *   (r, r+, w, w+, a, a+) see manpage for fopen for more info.
   */
  TFile (const TString& path, const TString& mode);
  
  /**
   * Closes any open file.
   */
  ~TFile ();
  
  /**
   * Close any open file.
   */
  void Close ();
  
  /**
   * Get address and length of mapped memory.
   *
   * @param start   Start of mapped memory.
   * @param length  Length of mapped memory.
   * @return        true, mapping exists or false, no mapping exists.
   */
  bool GetMap (void*& start, long& length);
  
  /**
   * Query if a file is open.
   *
   * @return  true, valid or false, not valid.
   */
  bool IsValid ();
  
  /**
   * Return the current length of the file.
   *
   * @return  Length of file or -1 on failure.
   */
  long Length ();
  
  /**
   * Map file into memory.
   * Reading or writing past the current length of the file it not
   * allowed. Also, mappging is only guaranteed to work if the file
   * has been opened with read only or read and write permissions.
   *
   * @return  Ptr to mapped memory or NULL on failure.
   */
  void * Map ();
  
  /**
   * Map file into memory.
   * Reading or writing past the current length of the file it not
   * allowed. Also, mappging is only guaranteed to work if the file
   * has been opened with read only or read and write permissions.
   *
   * @param offset  Offset into file to start map at.
   * @param lentgh  Length of region of file to map.
   * @return        Ptr to mapped memory or NULL on failure.
   */
  void * Map (long offset, long length);
  
  /**
   * Un map previously mapped memory.
   */
  void UnMap ();
  
  /**
   * Open the specified file.
   *
   * @param path  Path to file.
   * @param mode  Mode as passed to fopen().
   * @return      true, success or false, failure.
   */
  bool Open (const TString& path, const TString& mode);
 
  /**
   * Path name as recorded in the object regardless of its state.
   *
   * @return  Path name
   */
  inline TString Path () { return myPath; }
  
  /**
   * Read bytes from file.
   *
   * @param buf    Buffer to put data in.
   * @param count  Read this many bytes.
   * @return       Bytes read or -1 on failure.
   */
  long Read (void *buf, long count);
  
  /**
   * Seek in file.
   *
   * @param offset  Seek this many bytes-
   * @param whence  Starting position as in enum Seek.
   * @return        File position or -1 on seek error.
   */
  long Seek (long offset, enum Seek whence = SET);
  
  /**
   * Move all modified data and attributes of the file to the storage
   * device. This call blocks until the operation is completed.
   */
  void Sync ();

  /**
   * Write other file to file.
   * The file position of file is not rewinded.
   *
   * @param file   File to write.
   * @return       true, success or false, failure.
   */
  bool Write (const TFile& file);
  
  /**
   * Write string to file.
   *
   * @param str    String to write.
   * @return       true, success or false, failure.
   */
  bool Write (const TString& str);
  
  /**
   * Write data to file.
   *
   * @param buf    Ptr to source buffer.
   * @param count  Number of bytes to write.
   * @return       Number of bytes written or -1 on failure.
   */
  long Write (const void *buf, long count);
  
  
private:
  
  /** Path to file. */
  TString myPath;

  /** Pointer to mapped memory. */
  void *myMemory;
  
  /** Length of mapped memory. */
  long myMemoryLen;
  
#if defined (_WIN32) || defined (__CYGWIN__)
  /** File descriptor. */
  HANDLE myFildes;
  
  /** Handle for mapped memory. */
  HANDLE myMapdes;
  
  /** Protection used for mapped memory. */
  DWORD myProt;
  
#else
  /** File descriptor. */
  int myFildes;
  
  /** Protection used for mapped memory. */
  int myProt;
#endif
};

#endif // _T_FILE_HH
