#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include "TFile.hh"


TFile::TFile ()
  : myMemory (NULL),
    myFildes (-1),
    myProt (0)
{
}


TFile::TFile (const TString& path, const TString& mode)
  : myMemory (NULL),
    myFildes (-1),
    myProt (0)
{
  Open (path, mode);
}


TFile::~TFile ()
{
  Close ();
}


void
TFile::Close ()
{
  if (myFildes != -1)
    {
      UnMap ();
      close (myFildes);
      myFildes = -1;
    }
}


bool
TFile::GetMap (void*& start, long& length)
{
  start = myMemory;
  length = myMemoryLen;
  
  return myMemory;
}


bool
TFile::IsValid ()
{
  return (myFildes != -1);
}


long
TFile::Length ()
{
  long cur_pos, length = -1;
  
  if (myFildes != -1)
    {
      cur_pos = lseek (myFildes, 0, SEEK_CUR);
      length = lseek (myFildes, 0, SEEK_END);
      lseek (myFildes, cur_pos, SEEK_SET);
    }
  
  return length;
}


void *
TFile::Map ()
{
  long length = Length ();
  
  if (myFildes != -1 && myMemory == NULL && myProt && length > 0)
    {
      myMemoryLen = length;
      myMemory = mmap (0, length, myProt, MAP_SHARED, myFildes, 0);
      if (myMemory == MAP_FAILED)
	myMemory = NULL;
    }
  
  return myMemory;
}


void *
TFile::Map (long offset, long length)
{
  if (myFildes != -1 && myMemory == NULL && myProt && length > 0 &&
      offset > 0 && (offset + length) <= Length ())
    { 
      myMemoryLen = length;
      myMemory = mmap (0, length, myProt, MAP_SHARED, myFildes, offset);
      if (myMemory == MAP_FAILED)
	myMemory = NULL;
    }
  
  return myMemory;
}


void
TFile::UnMap ()
{
  if (myMemory)
    {
      munmap (myMemory, myMemoryLen);
      myMemory = NULL;
    }
}


bool
TFile::Open (const TString& path, const TString& mode)
{
  int flags = 0;
  bool flags_ok = true;
  
  myPath = path;
  
  if (myFildes == -1)
    {
      if (mode == "r")
	{
	  flags = O_RDONLY;
	  myProt = PROT_READ;
	}
      else if (mode == "r+")
	{
	  flags = O_RDWR;
	  myProt = PROT_READ | PROT_WRITE;
	}
      else if (mode == "w")
	{
	  flags = O_WRONLY | O_CREAT | O_TRUNC;
	  // Not possible to map write-only file
	  myProt = 0;
	}
      else if (mode == "w+")
	{
	  flags = O_RDWR | O_CREAT | O_TRUNC;
	  myProt = PROT_READ | PROT_WRITE;
	}
      else if (mode == "a")
	{
	  flags = O_WRONLY | O_CREAT | O_APPEND;
	  // Not possible to map write-only file
	  myProt = 0;
	}
      else if (mode == "a+")
	{
	  flags = O_RDWR | O_CREAT | O_APPEND;
	  myProt = PROT_READ | PROT_WRITE;
	}
      else
	flags_ok = false;
      
      if (flags_ok)
	myFildes = open ((const char *)path, flags, 0666);
    }
  
  return (myFildes != -1);
}


long
TFile::Read (void *buf, long count)
{
  if (myFildes == -1)
    return -1;
  
  return read (myFildes, buf, count);
}


long
TFile::Seek (long offset, enum Seek whence)
{
  int translate[3] = { SEEK_SET, SEEK_CUR, SEEK_END };
  
  if (myFildes == -1)
    return -1;
  
  return lseek (myFildes, offset, translate[whence]);
}


void
TFile::Sync ()
{
  if (myFildes != -1)
    {
      if (myMemory)
	msync (myMemory, myMemoryLen, MS_SYNC);
      
      fsync (myFildes);
    }
}


bool
TFile::Write (const TFile& file)
{
  long r;
  char buf[4096];
  
  if (myFildes == -1 || file.myFildes == -1)
    return false;
  
  while ((r = read (file.myFildes, buf, 4096)) > 0)
    {
      // FIXME: Add retry mechanism
      if (write (myFildes, buf, r) != r)
	return false;
    }
  
  if (r == -1)
    return false;
  
  return true;
}


bool
TFile::Write (const TString& str)
{
  int len = str.Length();
  
  if (myFildes == -1)
    return false;
  
  // FIXME: Add retry mechanism
  if (write (myFildes, str.Data(), len) != len)
    return false;
  
  return true;
}


long
TFile::Write (const void *buf, long count)
{
  if (myFildes == -1)
    return -1;
  
  return write (myFildes, buf, count);
}
