#include "TFile.hh"


TFile::TFile ()
  : myMemory (NULL),
    myFildes (INVALID_HANDLE_VALUE),
    myProt (0)
{
}


TFile::TFile (const TString& path, const TString& mode)
  : myMemory (NULL),
    myFildes (INVALID_HANDLE_VALUE),
    myProt (0)
{
  Open (path, mode);
}


TFile::~TFile ()
{
  Close ();
}


void
TFile::Close ()
{
  if (myFildes != INVALID_HANDLE_VALUE)
    {
      UnMap ();
      CloseHandle (myFildes);
      myFildes = INVALID_HANDLE_VALUE;
    }
}


bool
TFile::GetMap (void*& start, long& length)
{
  start = myMemory;
  length = myMemoryLen;
  
  return myMemory;
}


bool
TFile::IsValid ()
{
  return (myFildes != INVALID_HANDLE_VALUE);
}


long
TFile::Length ()
{
  long length = INVALID_FILE_SIZE;
  
  if (myFildes != INVALID_HANDLE_VALUE)
    {
      length = GetFileSize (myFildes, 0);
      if (length < 0)
	length = -1;
    }
  
  return length;
}


void *
TFile::Map ()
{
  long length = Length ();
  DWORD access;
  
  if (myFildes != INVALID_HANDLE_VALUE && myMemory == NULL &&
      myProt && length > 0)
    {
      myMapdes = CreateFileMapping (myFildes, 0, myProt, 0, 0, 0);
      if (! myMapdes)
	return myMemory;
      
      if (myProt == PAGE_READONLY)
	access = FILE_MAP_READ;
      else
	access = FILE_MAP_WRITE;
      
      myMemory = MapViewOfFile (myMapdes, access, 0, 0, length);
      myMemoryLen = length;
    }
  
  return myMemory;
}


void *
TFile::Map (long offset, long length)
{
  DWORD access;
  
  if (myFildes != INVALID_HANDLE_VALUE && myMemory == NULL && myProt &&
      length > 0 && offset > 0 && (offset + length) <= Length ())
    {
      myMapdes = CreateFileMapping (myFildes, 0, myProt, 0, 0, 0);
      if (! myMapdes)
	return myMemory;
      
      if (myProt == PAGE_READONLY)
	access = FILE_MAP_READ;
      else
	access = FILE_MAP_WRITE;
      
      myMemory = MapViewOfFile (myMapdes, access, 0, offset, length);
      myMemoryLen = length;
    }
  
  return myMemory;
}


void
TFile::UnMap ()
{
  if (myMemory)
    {
      UnmapViewOfFile (myMemory);
      CloseHandle (myMapdes);
      myMemory = NULL;
    }
}


bool
TFile::Open (const TString& path, const TString& mode)
{
  DWORD desired_access = 0, share_mode = 0, creation_disposition = 0;
  bool flags_ok = true, append = false;
  
  myPath = path;
  
  if (myFildes == INVALID_HANDLE_VALUE)
    {
      if (mode == "r")
	{
	  desired_access = GENERIC_READ;
	  creation_disposition = OPEN_EXISTING;
	  myProt = PAGE_READONLY;
	}
      else if (mode == "r+")
	{
	  desired_access = GENERIC_READ | GENERIC_WRITE;
	  creation_disposition = OPEN_EXISTING;
	  myProt = PAGE_READWRITE;
	}
      else if (mode == "w")
	{
	  desired_access = GENERIC_WRITE;
	  creation_disposition = CREATE_ALWAYS;
	  // Not possible to map write-only file
	  myProt = 0;
	}
      else if (mode == "w+")
	{
	  desired_access = GENERIC_READ | GENERIC_WRITE;
	  creation_disposition = CREATE_ALWAYS;
	  myProt = PAGE_READWRITE;
	}
      else if (mode == "a")
	{
	  desired_access = GENERIC_WRITE;
	  creation_disposition = OPEN_ALWAYS;
	  append = true;
	  // Not possible to map write-only file
	  myProt = 0;
	}
      else if (mode == "a+")
	{
	  desired_access = GENERIC_READ | GENERIC_WRITE;
	  creation_disposition = OPEN_ALWAYS;
	  append = true;
	  myProt = PAGE_READWRITE;
	}
      else
	flags_ok = false;
      
      if (flags_ok)
	myFildes = CreateFile ((const char *)path, desired_access,
			       share_mode, 0, creation_disposition,
			       FILE_ATTRIBUTE_NORMAL, 0);
      
      if (myFildes != INVALID_HANDLE_VALUE && append)
	Seek (0, TFile::END);
    }
  
  return (myFildes != INVALID_HANDLE_VALUE);
}


long
TFile::Read (void *buf, long count)
{
  if (myFildes == INVALID_HANDLE_VALUE || count < 0)
    return -1;
  
  DWORD bytes_read;
  
  if (! ReadFile (myFildes, buf, count, &bytes_read, 0))
    return -1;
  
  return bytes_read;
}


long
TFile::Seek (long offset, enum Seek whence)
{
  DWORD translate[3] = { FILE_BEGIN, FILE_CURRENT, FILE_END };
  DWORD position;
  
  if (myFildes == INVALID_HANDLE_VALUE)
    return -1;
  
  // INVALID_SET_FILE_POINTER
  position = SetFilePointer (myFildes, offset, 0, translate[whence]);
  if (position == INVALID_SET_FILE_POINTER)
    return -1;
  
  return position;
}


void
TFile::Sync ()
{
  if (myFildes != INVALID_HANDLE_VALUE)
    {
      if (myMemory)
	FlushViewOfFile (myMemory, 0);
      
      FlushFileBuffers (myFildes);
    }
}


bool
TFile::Write (const TFile& file)
{
  DWORD bytes_read, bytes_written;
  char buf[TFILE_BUF_LEN];
  
  if (myFildes == INVALID_HANDLE_VALUE ||
      file.myFildes == INVALID_HANDLE_VALUE)
    return false;
  
  do
    {
      if (! ReadFile (file.myFildes, buf, TFILE_BUF_LEN, &bytes_read, 0))
	return false;
      
      if (! WriteFile (myFildes, buf, bytes_read, &bytes_written, 0))
	return false;
      
      if (bytes_read != bytes_written)
	return false;
      
    } while (bytes_read == TFILE_BUF_LEN);
  
  return true;
}


bool
TFile::Write (const TString& str)
{
  if (myFildes == INVALID_HANDLE_VALUE)
    return false;
  
  DWORD bytes_written = 0;
  DWORD count = str.Length();
  
  WriteFile (myFildes, str.Data(), count, &bytes_written, 0);
  
  if (bytes_written != count)
    return false;
  
  return true;
}


long
TFile::Write (const void *buf, long count)
{
  DWORD bytes_written = 0;
  
  if (myFildes == INVALID_HANDLE_VALUE || count < 0)
    return -1;
  
  WriteFile (myFildes, buf, count, &bytes_written, 0);
  
  return bytes_written;
}
