#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TMapping.hh"
#include "TMemoryException.hh"

const int TMapping::BUCKET_MULTIPLIER = 3;
const int TMapping::NUM_BUCKETS = 4;


TMapping::TMapping ()
  : myNumBuckets (NUM_BUCKETS),
    myBucketMask (NUM_BUCKETS - 1),
    myGrowSize (NUM_BUCKETS * BUCKET_MULTIPLIER),
    myNumEntries (0)
{
  myBuckets = (TMappingEntry_t **) calloc (myNumBuckets, sizeof (myBuckets));
  if (! myBuckets)
    throw TMemoryException ();
}


TMapping::~TMapping ()
{
  Destruct ();
}


void
TMapping::Delete (TMappingEntry_t *entry)
{
  TMappingEntry_t *prev;
  int index;
  
  /* Only delete if ref_count becomes zero. */
  entry->ref_count--;
  if (entry->ref_count)
    return;
  
  index = entry->hash & myBucketMask;
  
  if (myBuckets[index] == entry)
    myBuckets[index] = entry->next;
  else
    {
      for (prev = myBuckets[index]; ; prev = prev->next)
	{
	  if (prev->next == entry)
	    {
	      prev->next = entry->next;
	      break;
	    }
	}
    }
  
  myNumEntries--;
  
  if (myNumBuckets > NUM_BUCKETS && myNumEntries <= myShrinkSize)
    Shrink ();
  
  free (entry->key);
  free (entry);  
}


TMappingEntry_t *
TMapping::Find (const char *cs, int n, unsigned int hash)
{
  TMappingEntry_t *entry;
  
  entry = myBuckets[hash & myBucketMask];
  while (entry)
    {
      if (entry->hash == hash)
	{
	  if (entry->bytes_used == n &&
	      TUtil::Equal (entry->key, cs, n))
	    return entry;
	}
      
      entry = entry->next;
    }
  
  return NULL;
}


TMappingEntry_t *
TMapping::Insert (const char *cs, int n, unsigned int hash)
{
  TMappingEntry_t *entry, *old_entry;
  int index, bytes_avail;
  
  if (! (entry = (TMappingEntry_t *) malloc (sizeof (TMappingEntry_t))))
    throw TMemoryException ();
  
  bytes_avail = MAPPING_KEY_SIZE(n);
  
  if (! (entry->key = (char *) malloc (bytes_avail)))
    throw TMemoryException ();
  
  /* Copy string and NULL terminate it. */
  memcpy (entry->key, cs, n);
  entry->key[n] = 0;
  
  /* Insert entry. */
  index = hash & myBucketMask;
  old_entry = myBuckets[index];
  myBuckets[index] = entry;
  
  entry->next = old_entry;
  entry->hash = hash;
  entry->bytes_used = n;
  entry->bytes_avail = bytes_avail;
  entry->ref_count = 1;
  
  myNumEntries++;
  if (myNumEntries >= myGrowSize)
    Grow ();
  
  return entry;
}


TMappingEntry_t *
TMapping::Insert (char *cs, int n1, int n2, unsigned int hash)
{
  TMappingEntry_t *entry, *old_entry;
  int index;
  
  if (! (entry = (TMappingEntry_t *) malloc (sizeof (TMappingEntry_t))))
    throw TMemoryException ();
  
  /* Insert entry. */
  index = hash & myBucketMask;
  old_entry = myBuckets[index];
  myBuckets[index] = entry;
  
  entry->next = old_entry;
  entry->key = cs;
  entry->hash = hash;
  entry->bytes_used = n1;
  entry->bytes_avail = n2;
  entry->ref_count = 1;
  
  myNumEntries++;
  if (myNumEntries >= myGrowSize)
    Grow ();
  
  return entry;
}


TMappingEntry_t *
TMapping::Link (TMappingEntry_t *entry)
{
  TMappingEntry_t *entry2;
  
  entry->hash = TUtil::Hash (entry->key, entry->bytes_used);
  
  entry2 = myBuckets[entry->hash & myBucketMask];
  while (entry2)
    {
      if (entry2->hash == entry->hash)
	{
	  if (entry2->bytes_used == entry->bytes_used &&
	      TUtil::Equal (entry2->key, entry->key, entry->bytes_used))
	    {
	      entry2->ref_count++;
	      
	      /* Destroy the input entry and return the found one. */	      
	      free (entry->key);
	      free (entry);
	      
	      return entry2;
	    }
	}
      
      entry2 = entry2->next;
    }
  
  /* Insert input entry */
  int index = entry->hash & myBucketMask;
  entry2 = myBuckets[index];
  myBuckets[index] = entry;
  
  entry->next = entry2;
  entry->ref_count = 1;
  
  myNumEntries++;
  if (myNumEntries >= myGrowSize)
    Grow ();
  
  return entry;
}


void
TMapping::Relocate (TMappingEntry_t *entry, unsigned int new_hash)
{
  TMappingEntry_t *prev, *old_entry;
  int index = entry->hash & myBucketMask;
  
  /* Unlink from current bucket. Assume that indexes are not the same,
     in which case we need not to relink. */
  if (myBuckets[index] == entry)
    myBuckets[index] = entry->next;
  else
    {
      for (prev = myBuckets[index]; ; prev = prev->next)
	{
	  if (prev->next == entry)
	    {
	      prev->next = entry->next;
	      break;
	    }
	}
    }
  
  /* Link to new bucket. */
  index = new_hash & myBucketMask;
  old_entry = myBuckets[index];
  myBuckets[index] = entry;
  entry->next = old_entry;
  entry->hash = new_hash;
}


void
TMapping::PrintStatus ()
{
  TMappingEntry_t *entry;
  int i, j;
  
  printf ("Number of buckets:  %d\n", myNumBuckets);
  printf ("Bucket mask:        %d\n", myBucketMask);
  printf ("Shrink size:        %d\n", myShrinkSize);
  printf ("Grow size:          %d\n", myGrowSize);
  printf ("Number of entries:  %d\n", myNumEntries);
  printf ("------------------\n");
  
  for (i = 0; i < myNumBuckets; i++)
    {
      printf ("Bucket: %d\n", i);
      
      j = 0;
      entry = myBuckets[i];
      while (entry)
	{
	  printf ("  Entry: %d\n", j);
	  printf ("    Key: %s\n", entry->key);
	  printf ("    Hash: %x\n", entry->hash);
	  printf ("    Bytes used: %d\n", entry->bytes_used);
	  printf ("    Bytes avail: %d\n", entry->bytes_avail);
	  printf ("    Ref count: %d\n", entry->ref_count);
	  
	  j++;
	  entry = entry->next;
	}
    }
}


TMappingEntry_t *
TMapping::Unlink (TMappingEntry_t *entry)
{
  entry->ref_count--;
  
  /* Make copy if used by more than one object */
  if (entry->ref_count)
    {
      TMappingEntry_t *entry2;
      
      if (! (entry2 = (TMappingEntry_t *) malloc (sizeof (TMappingEntry_t))))
	throw TMemoryException ();
      
      if (! (entry2->key = (char *) malloc (entry->bytes_avail)))
	throw TMemoryException ();
      
      entry2->hash = entry->hash;
      entry2->bytes_used = entry->bytes_used;
      entry2->bytes_avail = entry->bytes_avail;
      
      memcpy (entry2->key, entry->key, entry->bytes_used + 1);
      
      return entry2;
    }
  
  TMappingEntry_t *prev;
  int index = entry->hash & myBucketMask;
  
  if (myBuckets[index] == entry)
    myBuckets[index] = entry->next;
  else
    {
      for (prev = myBuckets[index]; ; prev = prev->next)
	{
	  if (prev->next == entry)
	    {
	      prev->next = entry->next;
	      break;
	    }
	}
    }
  
  myNumEntries--;
  
  if (myNumBuckets > NUM_BUCKETS && myNumEntries <= myShrinkSize)
    Shrink ();
  
  return entry;
}


void
TMapping::Destruct ()
{
  TMappingEntry_t *entry, *t;
  int i;
  
  if (myNumEntries)
    {
      for (i = 0; i < myNumBuckets; i++)
	{
	  entry = myBuckets[i];
	  while (entry)
	    {
	      t = entry;
	      entry = entry->next;
	      free (t->key);
	      free (t);
	    }
	}
      
      free (myBuckets);
      myNumEntries = 0;
    }
}


void
TMapping::Grow ()
{
  ReHash (myNumBuckets << 2);
  myShrinkSize = (myNumBuckets * BUCKET_MULTIPLIER) >> 4;
  myGrowSize <<= 2;
}


void
TMapping::ReHash (int new_num_buckets)
{
  TMappingEntry_t **orig_buckets = myBuckets;
  TMappingEntry_t *entry, *old_entry, *tmp_entry;
  int old_num_buckets = myNumBuckets, index;
  
  myBuckets = (TMappingEntry_t **)
    calloc (new_num_buckets, sizeof (myBuckets));
  
  if (! myBuckets)
    throw TMemoryException ();
  
  myNumBuckets = new_num_buckets;
  myBucketMask = new_num_buckets - 1;
  
  for (int i = 0; i < old_num_buckets; i++)
    {
      entry = orig_buckets[i];
      while (entry)
	{
	  tmp_entry = entry;
	  entry = entry->next;
	  index = tmp_entry->hash & myBucketMask;
	  old_entry = myBuckets[index];
	  myBuckets[index] = tmp_entry;
	  tmp_entry->next = old_entry;	  
	}
    }
  
  free (orig_buckets);
}


void
TMapping::Shrink ()
{
  ReHash (myNumBuckets >> 2);
  myShrinkSize >>= 2;
  myGrowSize >>= 2;
}
