#ifndef _T_MAPPING_HH
#define _T_MAPPING_HH

#include "TUtil.hh"
#include "TSingletonDestroyerT.hh"


/* Used to calculate the number of bytes to allocate for keys.
   n is the requested number of bytes. This always gives room for a
   trailing null character and the block size is 8 bytes. Some extra
   space is added for really big keys. */

#define MAPPING_KEY_SIZE(n) ((n + (n >> 4) + 15) & ~7)


typedef struct TMappingEntry TMappingEntry_t;

/**
 * A mapping contains entries like this.
 * The next and hash members should not be changed by other classes.
 * Use RelocateEntry() to update the entry with a new hash value.
 */

struct TMappingEntry
{
  /* Pointer to next entry in bucket or NULL for end of chain. */
  TMappingEntry_t *next;
  
  char *key;            /* Key (malloced memory). */
  unsigned int hash;    /* Hash value for key. */
  int bytes_used;       /* Number of bytes used. */
  int bytes_avail;      /* Number of bytes available. */
  int ref_count;        /* Reference count. */
};


/**
 * A mapping is much like a hash table except that only keys are used,
 * no associated values.
 *
 * Important: This class is only to be used by TString. Use
 * TValHasMapT or similar for hash maps.
 *
 * @author Johan Bolmsjo <johan@nocrew.org>
 */

class TMapping
{
public:
  
  /**
   * Delete entry.
   *
   * @param entry  Mapping entry to delete.
   */
  void Delete (TMappingEntry_t *entry);
  
  /**
   * Search for entry matching the specified key.
   *
   * @param cs    Ptr to key.
   * @param n     Length of key.
   * @param hash  Pre calculated has value.
   * @return      Ptr to found entry or NULL if none found.
   */
  TMappingEntry_t* Find (const char *cs, int n, unsigned int hash);
  
  /**
   * Insert entry according to the specified key.
   * Memory for key is allocated and the content of cs is copied.
   *
   * @param cs    Ptr to key.
   * @param n     Length of key.
   * @param hash  Pre calculated hash value.
   * @return      Ptr to inserted entry.
   */
  TMappingEntry_t* Insert (const char *cs, int n, unsigned int hash);
  
  /**
   * Insert entry according to the specified key.
   * The pointer cs is used as the key. It must have been allocated
   * with malloc.
   *
   * @param cs    Ptr to key.
   * @param n1    Length of key.
   * @param n2    Length of allocated memory.
   * @param hash  Pre calculated hash value.
   * @return      Ptr to inserted entry or NULL on error.
   */
  TMappingEntry_t* Insert (char *cs, int n1, int n2, unsigned int hash);
  
  /**
   * Link the specified entry.
   * If the entry is already in the mapping, that one is returned and the
   * ref count for it increased.
   *
   * @param entry  Entry to link
   * @return       Linked entry (or the one found)
   */
  TMappingEntry_t* Link (TMappingEntry_t *entry);
  
  /**
   * Prints status about the mapping to stdout.
   */
  void PrintStatus ();
  
  /**
   * Relocate mapping entry to a different bucket if the content
   * has been changed by some other class.
   *
   * @param entry     Mapping entry to relocate.
   * @param new_hash  New hash value for key.
   */
  void Relocate (TMappingEntry_t *entry, unsigned int new_hash);
  
  /**
   * Unlink the specified entry.
   * If the reference count is greater than 1 a copy is made and
   * returned.
   *
   * @param entry  Entry to unlink
   * @return       Unlinked entry (or copy)
   */
  TMappingEntry_t* Unlink (TMappingEntry_t *entry);
  
  
protected:

  /*  Only provide limited access. */
  friend class TString;
  friend class TSingletonDestroyerT<TMapping>;
  
  TMapping ();
  
  ~TMapping ();
  
  
private:
  
  /* FIXME: No copying for now! */
  TMapping (const TMapping&);
  
  /* FIXME: No copying for now! */
  TMapping& operator=(const TMapping&);
  
  /**
   * Grow bucket array.
   */
  void Grow ();
  
  /**
   * Re-bash bucket array to new number of buckets.
   * Updates myNumBuckets and myBucketMask.
   *
   * @param new_num_buckets  New number of buckets.
   */
  void ReHash (int new_num_buckets);
  
  /**
   * Shrink bucket array.
   */
  void Shrink ();
  
  /**
   * Free allocated resources.
   */
  void Destruct ();
  
  
  /* Ptr array to buckets. */
  TMappingEntry_t **myBuckets;
  
  /* Number of buckets in ptr array. */
  int myNumBuckets;
  
  /* Mask to apply to hash value to find the correct bucket. */
  int myBucketMask;

  /* Shrink table when the number of keys are less than or equal
     to this value. */
  int myShrinkSize;
  
  /* Grow table when the number of keys are greater than or equal
      to this value. */
  int myGrowSize;
  
  /* This many entries are in the table. */
  int myNumEntries;
  
  /* When there are this many keys per bucket, on average, rebuild
      the table to make it larger. */
  static const int BUCKET_MULTIPLIER;
  
  /* Initial number of buckets (powers of two). */
  static const int NUM_BUCKETS;
};

#endif // _T_MAPPING_HH
