#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#if ! (defined (_WIN32) || defined (__CYGWIN__))
#include <libgen.h>
#endif
#include "TBuffer.hh"
#include "TString.hh"
#include "TBoundsException.hh"
#include "TMemoryException.hh"

// Saturate values to 64
#define MAX_64(n) (n > 64) ? 64 : n

TMapping* TString::myMapping = 0;
TMappingEntry_t* TString::myNullString = 0;
TSingletonDestroyerT<TMapping> TString::myMappingDestroyer;


TString::TString ()
{
  ConstructMapping ();
  myMappingEntry = myNullString;
  myMappingEntry->ref_count++;
}


TString::TString (const TString& str)
{
  myMappingEntry = str.myMappingEntry;
  myMappingEntry->ref_count++;
}


TString::TString (const char *cs)
{
  ConstructMapping ();
  GetEntry (cs, Length (cs));
}


TString::TString (const char *cs, int n)
{
  ConstructMapping ();
  GetEntry (cs, n);
}


TString::TString (char c)
{
  ConstructMapping ();
  GetEntry (&c, 1);
}


TString::TString (char c, int n)
{
  char *cs;
  int bytes_avail = MAPPING_KEY_SIZE(n);
  
  if (! (cs = (char *) malloc (bytes_avail)))
    throw TMemoryException ();
  
  for (int i = 0; i < n; i++)
    cs[i] = c;
  
  /* Pre calculate hash value as we will be using it two times. */
  unsigned int hash = TUtil::Hash (cs, n);
  
  ConstructMapping ();
  myMappingEntry = myMapping->Find (cs, n, hash);
  if (myMappingEntry)
    {
      /* Return memory if this string already exist. */
      free (cs);
      myMappingEntry->ref_count++;
    }
  else
    {
      /* Null terminate string. */
      cs[n] = 0;
      myMappingEntry = myMapping->Insert (cs, n, bytes_avail, hash);
    }
}


TString::~TString ()
{
  /* Decrements and tests ref_count before deleting. */
  myMapping->Delete (myMappingEntry);
}


TString&
TString::operator=(const char* cs)
{
  int n = Length (cs);
  
  /* Release current entry. */
  myMapping->Delete (myMappingEntry);
  
  /* Acquire new entry. */
  GetEntry (cs, n);
  return *this;
}


TString&
TString::operator=(const TString& str)
{
  /* Only copy if different objects. */
  if (&str != this)
    {
      /* Release current entry. */
      myMapping->Delete (myMappingEntry);
      
      /* Acquire new entry. */
      myMappingEntry = str.myMappingEntry;
      myMappingEntry->ref_count++;
    }
  
  return *this;
}


// FIXME: optimise!
TString&
TString::Append (char c, int n)
{
  char buf[64];
  int buf_sz = MAX_64(n);
  
  // memset faster for longer strings?
  for (int i = 0; i < buf_sz; i++)
    buf[i] = c;
  
  while (n > 0)
    {
      Append (buf, MAX_64(n));
      n -= 64;
    }
  
  return *this;
}


TString&
TString::Append (const char* cs, int n)
{
  char *key;
  TMappingEntry_t *entry;
  unsigned int hash;
  int bytes_avail, bytes_used = myMappingEntry->bytes_used + n;
  
  // Can't append empty strings
  if (! n)
    return *this;
  
  if (myMappingEntry->ref_count == 1)
    {
      // Check if we can reuse already allocated memory
      if (bytes_used < myMappingEntry->bytes_avail)
	{
	  key = myMappingEntry->key;
	  memcpy (key + myMappingEntry->bytes_used, cs, n);
	  
	  // Check if string already exists
	  hash = TUtil::Hash (cs, n, myMappingEntry->hash);
	  entry = myMapping->Find (key, bytes_used, hash);
	  if (entry)
	    {
	      myMapping->Delete (myMappingEntry);
	      myMappingEntry = entry;
	      entry->ref_count++;
	    }
	  else
	    {
	      myMappingEntry->key[bytes_used] = 0;
	      myMappingEntry->bytes_used = bytes_used;
	      myMapping->Relocate (myMappingEntry, hash);
	    }
	}
      else
	{
	  bytes_avail = MAPPING_KEY_SIZE (bytes_used);
	  key = (char *) realloc (myMappingEntry->key, bytes_avail);
	  if (! key)
	    throw TMemoryException ();
	  
	  // Only update key here, needed by Delete and Find
	  myMappingEntry->key = key;
	  memcpy (key + myMappingEntry->bytes_used, cs, n);
	  
	  // Check if string already exists
	  hash = TUtil::Hash (cs, n, myMappingEntry->hash);
	  entry = myMapping->Find (key, bytes_used, hash);
	  if (entry)
	    {
	      myMapping->Delete (myMappingEntry);
	      myMappingEntry = entry;
	      entry->ref_count++;
	    }
	  else
	    {
	      myMappingEntry->bytes_used = bytes_used;
	      myMappingEntry->bytes_avail = bytes_avail;
	      key[bytes_used] = 0;
	      myMapping->Relocate (myMappingEntry, hash);
	    }
	}
    }
  else
    {
      bytes_avail = MAPPING_KEY_SIZE (bytes_used);
      if (! (key = (char *) malloc (bytes_avail)))
	throw TMemoryException ();
      
      memcpy (key, myMappingEntry->key, myMappingEntry->bytes_used);
      memcpy (key + myMappingEntry->bytes_used, cs, n);
      
      myMapping->Delete (myMappingEntry);
      
      // Check if string already exists
      hash = TUtil::Hash (cs, n, myMappingEntry->hash);
      entry = myMapping->Find (key, bytes_used, hash);
      if (entry)
	{
	  myMappingEntry = entry;
	  entry->ref_count++;
	  free (key);
	}
      else
	{
	  // Null terminate string
	  key[bytes_used] = 0;
	  myMappingEntry = myMapping->Insert (key, bytes_used,
					       bytes_avail, hash);
	}
    }
  
  return *this;
}


TString&
TString::Clear ()
{
  myMapping->Delete (myMappingEntry);
  myMappingEntry = myNullString;
  myMappingEntry->ref_count++;
  
  return *this;
}


TString&
TString::Insert (int pos, char c, int n)
{
  char *key;
  TMappingEntry_t *entry;
  unsigned int hash;
  int bytes_avail, bytes_used = myMappingEntry->bytes_used + n;
  
  // Ignore empty inserts
  if (! n)
    return *this;
  
#ifdef TOOLS_BOUND_CHECK
  if (pos < 0 || pos > myMappingEntry->bytes_used)
    throw TBoundsException ();
#endif
  
  bytes_avail = MAPPING_KEY_SIZE (bytes_used);
  if (! (key = (char *) malloc (bytes_avail)))
    throw TMemoryException ();
  
  memcpy (key, myMappingEntry->key, pos);
  memset (key + pos, c, n);
  memcpy (key + pos + n, myMappingEntry->key + pos,
	  myMappingEntry->bytes_used - pos);
  
  myMapping->Delete (myMappingEntry);
  
  // Check if string already exists
  hash = TUtil::Hash (key, bytes_used);
  entry = myMapping->Find (key, bytes_used, hash);
  if (entry)
    {
      myMappingEntry = entry;
      entry->ref_count++;
      free (key);
    }
  else
    {
      key[bytes_used] = 0;
      myMappingEntry = myMapping->Insert (key, bytes_used,
					  bytes_avail, hash);
    }
  
  return *this;
}


TString&
TString::Insert (int pos, const char *cs, int n) 
{
  char *key;
  TMappingEntry_t *entry;
  unsigned int hash;
  int bytes_avail, bytes_used = myMappingEntry->bytes_used + n;
  
  // Ignore empty inserts
  if (! n)
    return *this;
  
#ifdef TOOLS_BOUND_CHECK
  if (pos < 0 || pos > myMappingEntry->bytes_used)
    throw TBoundsException ();
#endif
  
  bytes_avail = MAPPING_KEY_SIZE (bytes_used);
  if (! (key = (char *) malloc (bytes_avail)))
    throw TMemoryException ();
  
  memcpy (key, myMappingEntry->key, pos);
  memcpy (key + pos, cs, n);
  memcpy (key + pos + n, myMappingEntry->key + pos,
	  myMappingEntry->bytes_used - pos);
  
  myMapping->Delete (myMappingEntry);
  
  // Check if string already exists
  hash = TUtil::Hash (key, bytes_used);
  entry = myMapping->Find (key, bytes_used, hash);
  if (entry)
    {
      myMappingEntry = entry;
      entry->ref_count++;
      free (key);
    }
  else
    {
      key[bytes_used] = 0;
      myMappingEntry = myMapping->Insert (key, bytes_used,
					  bytes_avail, hash);
    }
  
  return *this;
}


TString&
TString::Format (const char *fmt, ...)
{
  TBuffer tbuf;
  int len = 0, coff, i;
  unsigned int u;
  char cbuf[16], c, *s;
  const char *p = fmt;
  bool magic = false, negative;
  
  va_list ap;
  va_start (ap, fmt);
  
  while ((c = *fmt++))
    {
      if (magic)
	{
	  switch (c)
	    {
	    case '%':
	      len++;
	      continue;
	      
	    // Signed integer
	    case 'd':
	      magic = false;
	      if (len)
		{
		  tbuf.Append (p, len);
		  len = 0;
		}
	      
	      p = fmt;
	      i = va_arg (ap, int);
	      if (i < 0)
		{
		  u = -i;
		  negative = true;
		}
	      else
		{
		  u = i;
		  negative = false;
		}
	      
	      coff = 16;
	      do
		{
		  cbuf[--coff] = '0' + u % 10;
		  u /= 10;
		} while (u);
	      
	      if (negative)
		cbuf[--coff] = '-';
	      
	      tbuf.Append (cbuf + coff, 16 - coff);
	      continue;
	    
	    // Unsigned integer
	    case 'u':
	      magic = false;
	      if (len)
		{
		  tbuf.Append (p, len);
		  len = 0;
		}
	      
	      p = fmt;
	      u = va_arg (ap, unsigned int);
	      coff = 16;
	      
	      do
		{
		  cbuf[--coff] = '0' + u % 10;
		  u /= 10;
		} while (u);
	      
	      tbuf.Append (cbuf + coff, 16 - coff);
	      continue;
	      
	    // Null terminated string
	    case 's':
	      magic = false;
	      if (len)
		{
		  tbuf.Append (p, len);
		  len = 0;
		}
	      p = fmt;
	      s = va_arg (ap, char *);
	      tbuf.Append (s);
	      continue;
	      
	    default:
	      magic = false;
	      len++;
	      continue;
	    }
	}
      
      if (c == '%')
	magic = true;
      else
	len++;
    }
  
  va_end (ap);
  
  if (len)
    tbuf.Append (p, len);
  else if (magic)
    tbuf.Append (p, 1);
  
  myMapping->Delete (myMappingEntry);
  GetEntry (tbuf.Data (), tbuf.Length ());
  
  return *this;
}


bool
TString::IsAscii () const
{
  for (int i = 0; i < myMappingEntry->bytes_used; i++)
    if (myMappingEntry->key[i] & 0x80)
      return false;
  
  return true;
}


bool
TString::IsDigits () const
{
  int c;
  
  for (int i = 0; i < myMappingEntry->bytes_used; i++)
    {
      c = myMappingEntry->key[i];
      if (c < '0' || c > '9')
	return false;
    }
  
  return true;
}


bool
TString::IsXDigits () const
{
  int c;
  
  for (int i = 0; i < myMappingEntry->bytes_used; i++)
    {
      c = myMappingEntry->key[i];
      if ((c < '0' || c > '9') &&
	  (c < 'A' || c > 'F') &&
	  (c < 'a' || c > 'f'))
	return false;
    }
  
  return true;
}


// FIXME: write generic basename and dirname functions
#if ! (defined (_WIN32) || defined (__CYGWIN__))
TString
TString::BaseName () const
{
  char *name = basename (myMappingEntry->key);
  return TString (name);
}


TString
TString::DirName () const
{
  char *name = dirname (myMappingEntry->key);
  return TString (name);
}
#endif


int
TString::Integer () const
{
  char *cs = myMappingEntry->key;
  int i = 0, v = 0, l = myMappingEntry->bytes_used;
  unsigned int t;
  bool negate = false;
  
  /* Check for - at first position. */
  if (l && *cs == '-')
    {
      negate = true;
      i++;
    }
  
  for (; i < l; i++)
    {
      t = cs[i] - '0';
      if (t <= 9)
	v = v * 10 + t;
      else
	break;
    }
  
  if (negate)
    v = -v;
  
  return v;
}


TString&
TString::Lower ()
{
  TMappingEntry_t *entry = myMapping->Unlink (myMappingEntry);
  int c;
  
  for (int i = 0; i < entry->bytes_used; i++)
    {
      c = entry->key[i];
      if (c >= 'A' && c <= 'Z')
	entry->key[i] = c + 32;
    }
  
  myMappingEntry = myMapping->Link (entry);
  return *this;
}


void
TString::Print (bool newline) const
{
  if (newline)
    printf ("%s\n", myMappingEntry->key);
  else
    printf ("%s", myMappingEntry->key);
}


// Does not shrink string
TString&
TString::Trim ()
{
  int t, left = 0, right = myMappingEntry->bytes_used - 1;
  
  for (; left <= right; left++)
    {
      t = myMappingEntry->key[left];
      if (t != ' ' && t != '\t')
	break;
    }
  
  for (; right >= left; right--)
    {
      t = myMappingEntry->key[right];
      if (t != ' ' && t != '\t')
	break;
    }
  
  t = right - left + 1;
  if (t != myMappingEntry->bytes_used)
    {
      TMappingEntry_t *entry = myMapping->Unlink (myMappingEntry);
      
      for (int i = 0; i < t; left++, i++)
	entry->key[i] = entry->key[left];
      
      entry->key[t] = 0;
      entry->bytes_used = t;
      
      myMappingEntry = myMapping->Link (entry);
    }
  
  return *this;
}


// Does not shrink string
TString&
TString::TrimLeft ()
{
  int t, left;
  
  for (left = 0; left < myMappingEntry->bytes_used; left++)
    {
      t = myMappingEntry->key[left];
      if (t != ' ' && t != '\t')
	break;
    }
  
  t = myMappingEntry->bytes_used - left;
  if (t != myMappingEntry->bytes_used)
    {
      TMappingEntry_t *entry = myMapping->Unlink (myMappingEntry);
      
      for (int i = 0; i < t; left++, i++)
	entry->key[i] = entry->key[left];
      
      entry->key[t] = 0;
      entry->bytes_used = t;
      
      myMappingEntry = myMapping->Link (entry);
    }
  
  return *this;
}


// Does not shrink string
TString&
TString::TrimRight ()
{
  int t, right = myMappingEntry->bytes_used - 1;
  
  for (; right >= 0; right--)
    {
      t = myMappingEntry->key[right];
      if (t != ' ' && t != '\t')
	break;
    }
  
  t = right + 1;
  if (t != myMappingEntry->bytes_used)
    {
      TMappingEntry_t *entry = myMapping->Unlink (myMappingEntry);
      
      for (int i = 0; i < t; i++)
	entry->key[i] = entry->key[i];
      
      entry->key[t] = 0;
      entry->bytes_used = t;
      
      myMappingEntry = myMapping->Link (entry);
    }
  
  return *this;
}


// Does not shrink string
TString&
TString::Truncate (int pos)
{
  if (pos >= 0 && pos < myMappingEntry->bytes_used)
    {
      TMappingEntry_t *entry = myMapping->Unlink (myMappingEntry);
      
      entry->key[pos] = 0;
      entry->bytes_used = pos;
      
      myMappingEntry = myMapping->Link (entry);
    }
  
  return *this;
}


unsigned int
TString::UnsignedInteger () const
{
  char *cs = myMappingEntry->key;
  int v = 0, l = myMappingEntry->bytes_used;
  unsigned int t;
  
  for (int i = 0; i < l; i++)
    {
      t = cs[i] - '0';
      if (t <= 9)
	v = v * 10 + t;
      else
	break;
    }
  
  return v;
}


TString&
TString::Upper ()
{
  TMappingEntry_t *entry = myMapping->Unlink (myMappingEntry);
  int c;
  
  for (int i = 0; i < entry->bytes_used; i++)
    {
      c = entry->key[i];
      if (c >= 'a' && c <= 'z')
	entry->key[i] = c - 32;
    }
  
  myMappingEntry = myMapping->Link (entry);
  return *this;
}


// Private
void
TString::ConstructMapping ()
{
  static const char *empty_str = "";

  if (! myMapping)
    {
      myMapping = new TMapping ();
      myMappingDestroyer.SetObject (myMapping);
    }
  // Create null string to speed up default constructor
  if (! myNullString)
    {
      unsigned int hash = TUtil::Hash (empty_str, 0);
      
      myNullString = myMapping->Find (empty_str, 0, hash);
      if (myNullString)
	myNullString->ref_count++;
      else
	myNullString = myMapping->Insert (empty_str, 0, hash);
    }
}
