#ifndef _T_STRING_HH
#define _T_STRING_HH

#include <stdlib.h>
#include <string.h>
#include "TMapping.hh"
#include "TSingletonDestroyerT.hh"

/**
 * String class featuring unique strings.
 * String compares are extremely fast, modifications on the other hand
 * are slow, append is a special case wich is optimised.
 *
 * @author Johan Bolmsjo <johan@nocrew.org>
 */

class TString
{
public:
  
  /**
   * Constructs a string of length zero.
   */
  TString ();
  
  /**
   * Constructs a string from the null terminated string cs.
   */
  TString (const char *cs);
  
  /**
   * Constructs a string from the string cs of length n.
   */
  TString (const char *cs, int n);
  
  /**
   * Constructs a string from str.
   */
  TString (const TString& str);
  
  /**
   * Constructs a string containing the single character c.
   */
  TString (char c);
  
  /**
   * Constructs a string containing the character c repeated n times.
   */
  TString (char c, int n);
  
  /**
   * Destructor.
   */
  ~TString ();
  
  /**
   * Return internal data as a null terminated character string.
   * The data may not be modified and the pointer will not be
   * vaild once the string goes out of scope.
   */
  inline operator const char*() const { return myMappingEntry->key; }
  
  /**
   * Convert the string to an integer.
   */
  inline operator int() const { return Integer (); }
  
  /**
   * Convert the string to an unsigned integer.
   */
  inline operator unsigned int() const { return UnsignedInteger (); }
  
  /**
   * Copy the null terminated string cs.
   */
  TString& operator=(const char* cs);
  
  /**
   * Copy the data of string str.
   */
  TString& operator=(const TString& str);
  
  /**
   * Compare string for equality with other TString object.
   * This is a matter of comparing two pointers and is very fast!
   */
  inline bool operator==(const TString& str) const
  {
    return myMappingEntry == str.myMappingEntry;
  }
  
  /**
   * Compare string for equality with character string.
   */
  inline bool operator==(const char *cs) const
  {
    return ! strcmp (myMappingEntry->key, cs);
  }
  
  /**
   * Compares two string objects for non-equality.
   * This is a matter of comparing two pointers and is very fast!
   */
  inline bool operator!=(const TString& str)
  {
    return myMappingEntry != str.myMappingEntry;
  }
  
  /**
   * Append cs to string.
   */
  inline TString& operator+=(const char* cs)
  {
    return Append (cs, Length (cs));
  }
  
  /**
   * Append str to string.
   */
  inline TString& operator+=(const TString& str)
  {
    return Append (str.myMappingEntry->key, str.myMappingEntry->bytes_used);
  }
  
  /**
   * Append the character c to string.
   */
  inline TString& operator+=(char c)
  {
    return Append (&c, 1);
  }
  
  /**
   * Append two strings together creating a third.
   */
  inline TString operator+(const char *cs)
  {
    TString t (*this);
    t.Append (cs, Length (cs));
    return t;
  }
  
  /**
   * Append two strings together creating a third.
   */
  inline TString operator+(const TString& str)
  {
    TString t (*this);
    t.Append (str);
    return t;
  }
  
  /**
   * Index operator.
   * Provides read-only access only.
   */
  inline char operator[](int i) const
  {
    return ((char *)myMappingEntry->key)[i];
  }
  
  /**
   * Append the character c.
   */
  inline TString& Append (char c)
  {
    return Append (&c, 1);
  }
  
  /**
   * Append n copies of the character c.
   */
  TString& Append (char c, int n);

  /**
   * Append a copy of the null terminated string cs.
   */
  inline TString& Append (const char* cs)
  {
    return Append (cs, Length (cs));
  }
  
  /**
   * Append a copy of the first n bytes of cs.
   */
  TString& Append (const char* cs, int n);
  
  /**
   * Append a copy of str to string.
   */
  inline TString& Append (const TString& str)
  {
    return Append (str.myMappingEntry->key, str.myMappingEntry->bytes_used);
  }

#if ! (defined (_WIN32) || defined (__CYGWIN__))  
  /**
   * Return a string containing the basename of the string.
   * See the Unix man page for basename.
   */
  TString BaseName () const;
  
  /**
   * Return a string containing the dirname of the string.
   * See the Unix man page for dirname.
   */
  TString DirName () const;
#endif

  /**
   * Return the data of ths string as a null terminated character string.
   * The data may not be modified and the pointer will not be vaild once
   * the string goes out of scope.
   */
  inline const char *Data () const { return myMappingEntry->key; }

  /**
   * Clear string.
   */
  TString& Clear ();
  
  /**
   * Check if the string is empty.
   *
   * @return  true, empty or false populated
   */
  inline bool Empty () const { return !myMappingEntry->bytes_used; }
  
  /**
   * Format a string as printf does.
   * For now only the convertion codes %d, %u and %s are supported
   * (without additional formating).
   *
   * @param fmt  Format string, for example "Celivin is %d years old."
   * @param ...  Aditional arguments
   */
  TString& Format (const char *fmt, ...);
  
  /**
   * Insert the character c at position pos.
   */
  inline TString& Insert (int pos, char c)
  {
    return Insert (pos, c, 1);
  }
  
  /**
   * Insert n copies of the character c at position pos.
   */
  TString& Insert (int pos, char c, int n);
  
  /**
   * Insert a copy of the null terminated string cs at position pos.
   */
  inline TString& Insert (int pos, const char *cs)
  {
    return Insert (pos, cs, TString::Length (cs));
  }
  
  /**
   * Insert a copy of the first n bytes of cs at position pos.
   */
  TString& Insert (int pos, const char *cs, int n);
  
  /**
   * Insert a copy of str at position pos.
   */
  inline TString& Insert (int pos, const TString& str)
  {
    return Insert (pos, str.myMappingEntry->key,
		   str.myMappingEntry->bytes_used);
  }
  
  /**
   * Calculate hash value of string.
   *
   * @return  Hash value
   */
  inline unsigned int Hash () const { return myMappingEntry->hash; }
  
  /**
   * Check is the string conists of ASCII characters only.
   * That is bit 7 is cleared for all characters.
   *
   * @return  true, ASCII only or false, non ASCII
   */
  bool IsAscii () const;
  
  /**
   * Check if the string conists of decimal digits only.
   *
   * @return  true, false
   */
  bool IsDigits () const;
  
  /**
   * Check if the string conists of hexa decimal digits only.
   *
   * @return  true, false
   */
  bool IsXDigits () const;
  
  /**
   *  Returns the length of the string.
   */
  inline int Length () const { return myMappingEntry->bytes_used; }
  
  /**
   * Faster than strlen() for short strings and according to my tests
   * equal on long strings.
   *
   * @param cs  Null terminated C string.
   * @return    Length of string.
   */
  static inline int Length (const char *cs)
  {
    int n = 0;
    while (*cs++)
      n++;
    
    return n;
  }
  
  /**
   * Convert the string to an integer.
   * Valid characters are 0 to 9 and at the first position the '-'
   * character is also allowed. If an invalid character is found, what has
   * been converted so far is returned. For example the string 42e3 would
   * return an integer with the value of 42. There is no overflow check.
   *
   * @return  Integer representation of string.
   */
  int Integer () const;
  
  /**
   * Convert string to lowercase letters only.
   */
  TString& Lower ();
  
  /**
   * Prepend the character c.
   */
  inline TString& Prepend (char c)
  {
    return Insert (0, c, 1);
  }
  
  /**
   * Prepend n copies of the character c.
   */
  inline TString& Prepend (char c, int n)
  {
    return Insert (0, c, n);
  }
  
  /**
   * Prepend a copy of the null terminated string cs.
   */
  inline TString& Prepend (const char *cs)
  {
    return Insert (0, cs, TString::Length (cs));
  }
  
  /**
   * Prepend a copy of the first n bytes of cs.
   */
  inline TString& Prepend (const char *cs, int n)
  {
    return Insert (0, cs, n);
  }
  
  /**
   * Prepend a copy of str.
   */
  inline TString& Prepend (const TString& str)
  {
    return Insert (0, str.myMappingEntry->key,
		   str.myMappingEntry->bytes_used);
  }
  
  /**
   * Print string to stdout.
   * newline specifies if a newline should be printed after the string.
   */
  void Print (bool newline = true) const;
  
  /**
   * Print status from mapping.
   * Used to debug things.
   */
  static void PrintStatus ()
  {
    ConstructMapping ();
    myMapping->PrintStatus ();
  }
  
  /**
   * Trim whitespace from beginning and end of string.
   */
  TString& Trim ();
  
  /**
   * Trim whitespace from beginning of string.
   */
  TString& TrimLeft ();
  
  /**
   * Trim whitespace from end of string.
   */
  TString& TrimRight ();

  /**
   * Truncate string at the specified position.
   */
  TString& Truncate (int pos);
  
  /**
   * Convert the string to an unsigned integer.
   * Valid characters are 0 to 9. If an invalid character is found,
   * what has been converted so far is returned. For example the string
   * 42e3 would return an integer with the value of 42. There is no
   * overflow check.
   *
   * @return  Unsigned integer representation of string.
   */
  unsigned int UnsignedInteger () const;  
  
  /**
   * Convert string to uppercase letters only.
   */
  TString& Upper ();
  
  
private:
  
  /**
   * Convenience method used in most of the constructors.
   * Basicly it inserts a character string into the mapping.
   */
  inline void GetEntry (const char *cs, int n)
  {
    unsigned int hash = TUtil::Hash (cs, n);
    
    myMappingEntry = myMapping->Find (cs, n, hash);
    if (myMappingEntry)
      myMappingEntry->ref_count++;
    else
      myMappingEntry = myMapping->Insert (cs, n, hash);
  }
  
  /**
   * Construct mapping object and null string.
   */
  static void ConstructMapping ();
  
  /* Mapping used to store all strings in.
     Not an object because of static object construction order
     dependencies. */
  static TMapping* myMapping;
  
  /* Mapping entry for the NULL string. */
  static TMappingEntry_t* myNullString;
  
  /* Singleton destroyer to delete our mapping on program exit.
     Easier to spot real memory leaks if we clean up this way. */
  static TSingletonDestroyerT<TMapping> myMappingDestroyer;
  
  /* Mapping entry for our string. */
  TMappingEntry_t* myMappingEntry;
};

#endif // _T_STRING_HH
