#ifndef _T_VAL_HASH_MAP_T_HH
#define _T_VAL_HASH_MAP_T_HH

#include <stdio.h>
#include <stdlib.h>
#include "TMemoryException.hh"


// Template class used to hold one key, value pair in a TValHashMapT object.

template <class K, class V>
class TValHashMapEntryT
{
public:
  
  /**
   * Constructor to avoid call of constructor on key and value objects.
   */
  TValHashMapEntryT (const K& k, const V& v)
    : key (k),
      value (v)
  {
  }
  
  /** Pointer to next entry in bucket or NULL for end of chain. */
  TValHashMapEntryT *next;
  
  /** Hash value for key. */
  unsigned int hash;
  
  /** Key */
  K key;
  
  /** Value */
  V value;
};


/**
 * Template class that maintains a collection of keys (K), value (V)
 * pairs which are stored according to the hash utility class U.
 *
 * The class U should contain the static methods Hash and Equal to
 * calculate the hash of K objects and to compare K objects for
 * equality.
 *
 * @see  THashInt
 * @see  THashString
 * @see  THashT
 *
 * @author Johan Bolmsjo <johan@nocrew.org>
 */

template <class K, class V, class U>
class TValHashMapT
{
public:
  
  /**
   * Constructs an empty hash map.
   */
  TValHashMapT ();
  
  /**
   * Copy constructor.
   */
  inline TValHashMapT (const TValHashMapT<K,V,U> &map) { Copy (map); }
  
  /**
   * Destructor.
   */
  ~TValHashMapT ();
  
  /**
   * Applies the user defined function to every key, value pair in self.
   *
   * @param fn  Function pointer to your function.
   * @param d   Client data to pass to your function.
   */
  void Apply (void (*fn)(const K *, V *, void *), void *d);
  
  /**
   * Removes all keys from self.
   */
  void Clear ();
  
  /**
   * Check if there exists a key in self that compares equal to k.
   *
   * @param k  Key to look for.
   * @return   true, match or false, no match.
   */
  inline bool Contains (const K& k) { return Find (k); }
  
  /**
   * Query number of keys in self.
   *
   * @return  Number of keys.
   */
  inline int Entries () { return myNumEntries; }

  /**
   * Find value associated with key.
   * If the key is not found, v is not updated.
   *
   * @param k  Key to look for.
   * @return   true, match or false, no match.
   */
  bool FindValue (const K& k, V& v);
  
  /**
   * Insert a key, value pair to the collection.
   *
   * @param k  Key to insert.
   * @param v  Value to insert,
   * @return   true, key inserted or false, key already exists.
   */
  bool Insert (const K& k, const V& v);
  
  /**
   * Remove a key from the collection.
   *
   * @param k  Key to remove.
   * @return   true, key removed or false, no such key.
   */
  bool Remove (const K& k);
  
  /**
   * Enable or dissable shrinking of bucket array when keys are removed.
   * This saves memory but costs some CPU time. It is enabled by default.
   *
   * @param flag  true, enable or false, disable.
   */
  inline void Shrink (bool flag) { myIsShrinkEnabled = flag; }
  
  /**
   * Index operator.
   * Returns a reference to the value associated with the key k. The
   * value may be modified.
   */
  V& operator[](const K& k);
  
  /**
   * Assignment operator.
   */
  TValHashMapT<K,V,U>& operator=(const TValHashMapT<K,V,U> &map);
  
private:

  /**
   * Copy contents from other hash map.
   * Used by copy CTOR and assignment operator.
   */
  void Copy (const TValHashMapT<K,V,U>& map);
  
  /**
   * Find value associated with key.
   *
   * @param k  Key to compare with.
   * @return   Pointer to value or NULL if none found.
   */
  V* Find (const K& k);
  
  /**
   * Grow bucket array.
   */
  void Grow ();
  
  /**
   * Re-bash bucket array to new number of buckets.
   * Updates myNumBuckets and myBucketMask.
   *
   * @param new_num_buckets  New number of buckets.
   */
  void ReHash (int new_num_buckets);

  /**
   * Shrink bucket array.
   */
  void Shrink ();
  
  
  /** Buckets */
  TValHashMapEntryT<K,V> **myBuckets;
  
  /** Number of buckets in array. */
  int myNumBuckets;
  
  /** Mask to apply to hash value to find the correct bucket. */
  int myBucketMask;
  
  /** Shrink table when the number of keys are less than or equal
      to this value. */
  int myShrinkSize;
  
  /** Grow table when the number of keys are greater than or equal
      to this value. */
  int myGrowSize;
  
  /** This many entries are in the table. */
  int myNumEntries;
  
  /** Allow shrinking of bucket array. */
  bool myIsShrinkEnabled;
  
  /** When there are this many keys per bucket, on average, rebuild
      the table to make it larger. */
  static const int BUCKET_MULTIPLIER;
  
  /** Initial number of buckets (powers of two). */
  static const int NUM_BUCKETS;
};


template <class K, class V, class U>
const int TValHashMapT<K,V,U>::BUCKET_MULTIPLIER = 3;

template <class K, class V, class U>
const int TValHashMapT<K,V,U>::NUM_BUCKETS = 4;


template <class K, class V, class U>
TValHashMapT<K,V,U>::TValHashMapT ()
  : myNumBuckets (NUM_BUCKETS),
    myBucketMask (NUM_BUCKETS - 1),
    myGrowSize (NUM_BUCKETS * BUCKET_MULTIPLIER),
    myNumEntries (0),
    myIsShrinkEnabled (true)
{
  myBuckets = (TValHashMapEntryT<K,V> **)
    calloc (myNumBuckets, sizeof (myBuckets));
  
  if (! myBuckets)
    throw TMemoryException ();
}


template <class K, class V, class U>
TValHashMapT<K,V,U>::~TValHashMapT ()
{
  myIsShrinkEnabled = false;
  Clear ();
  free (myBuckets);
}


template <class K, class V, class U>
void TValHashMapT<K,V,U>::Apply (void (*fn)(const K *, V *, void *), void *d)
{
  TValHashMapEntryT<K,V> *entry;
  
  for (int i = 0; i < myNumBuckets; i++)
    {
      entry = myBuckets[i];
      while (entry)
	{
	  fn (&entry->key, &entry->value, d);
	  entry = entry->next;
	}
    }
}


template <class K, class V, class U>
void TValHashMapT<K,V,U>::Clear ()
{
  TValHashMapEntryT<K,V> *entry, *tmp_entry;
  
  if (myNumEntries)
    {
      for (int i = 0; i < myNumBuckets; i++)
	{
	  entry = myBuckets[i];
	  while (entry)
	    {
	      tmp_entry = entry;
	      entry = entry->next;
	      delete tmp_entry;
	    }
	  
	  myBuckets[i] = 0;
	}
      
      myNumEntries = 0;
    }
  
  if (myIsShrinkEnabled)
    {
      myNumBuckets = NUM_BUCKETS;
      myBucketMask = NUM_BUCKETS - 1;
      myGrowSize = NUM_BUCKETS * BUCKET_MULTIPLIER;
      
      myBuckets = (TValHashMapEntryT<K,V> **)
	realloc (myBuckets, myNumBuckets * sizeof (myBuckets));
    }
}


template <class K, class V, class U>
bool TValHashMapT<K,V,U>::FindValue (const K& k, V& v)
{
  V *value = Find (k);
  if (value)
    {
      v = *value;
      return true;
    }
  
  return false;
}


template <class K, class V, class U>
bool TValHashMapT<K,V,U>::Insert (const K& k, const V& v)
{
  TValHashMapEntryT<K,V> *entry, *old_entry;
  unsigned int hash = U::Hash (k);
  int index = hash & myBucketMask;
  
  entry = old_entry = myBuckets[index];
  while (entry)
    {
      if (entry->hash == hash &&
	  U::Equal (k, entry->key))
	return false;
      
      entry = entry->next;
    }
  
  entry = new TValHashMapEntryT<K,V> (k, v);
  myBuckets[index] = entry;
  entry->next = old_entry;
  entry->hash = hash;
  
  myNumEntries++;
  if (myNumEntries >= myGrowSize)
    Grow ();
  
  return true;
}


template <class K, class V, class U>
bool TValHashMapT<K,V,U>::Remove (const K& k)
{
  TValHashMapEntryT<K,V> *entry, *prev = 0;
  unsigned int hash = U::Hash (k);
  int index = hash & myBucketMask;
  
  entry = myBuckets[index];
  while (entry)
    {
      if (entry->hash == hash &&
	  U::Equal (k, entry->key))
	{
	  if (entry == myBuckets[index])
	    myBuckets[index] = entry->next;
	  else
	    prev->next = entry->next;
	  
	  delete (entry);
	  myNumEntries--;
	  if (myIsShrinkEnabled && myNumBuckets > NUM_BUCKETS &&
	      myNumEntries <= myShrinkSize)
	    Shrink ();
	  
	  return true;
	}
      
      prev = entry;
      entry = entry->next;
    }
  
  return false;
}


template <class K, class V, class U>
V& TValHashMapT<K,V,U>::operator[](const K& k)
{
  V *v = Find (k);
  
  if (v)
    return *v;
  else
    {
      V some_value;
      Insert (k, some_value);
      return *(Find (k));
    }
}


template <class K, class V, class U>
TValHashMapT<K,V,U>&
TValHashMapT<K,V,U>::operator=(const TValHashMapT<K,V,U> &map)
{
  /* Only copy if different objects. */
  if (&map != this)
    {
      bool shrink_enabled = myIsShrinkEnabled;
      
      myIsShrinkEnabled = false;
      Clear ();
      free (myBuckets);
      Copy (map);
      myIsShrinkEnabled = shrink_enabled;
    }
  
  return *this;
}


template <class K, class V, class U>
void TValHashMapT<K,V,U>::Copy (const TValHashMapT<K,V,U>& map)
{
  TValHashMapEntryT<K,V> *orig_entry, *new_entry, *old_entry;
  
  myNumBuckets = map.myNumBuckets;
  myBucketMask = map.myBucketMask;
  myShrinkSize = map.myShrinkSize;
  myGrowSize = map.myGrowSize;
  myNumEntries = map.myNumEntries;
  myIsShrinkEnabled = map.myIsShrinkEnabled;
  
  myBuckets = (TValHashMapEntryT<K,V> **)
    calloc (myNumBuckets, sizeof (myBuckets));
  
  if (! myBuckets)
    throw TMemoryException ();
  
  for (int i = 0; i < myNumBuckets; i++)
    {
      orig_entry = map.myBuckets[i];
      while (orig_entry)
	{
	  old_entry = myBuckets[i];
	  new_entry = new TValHashMapEntryT<K,V> (orig_entry->key,
						  orig_entry->value);
	  new_entry->next = old_entry;
	  new_entry->hash = orig_entry->hash;
	  myBuckets[i] = new_entry;
	  orig_entry = orig_entry->next;
	}
    }
}


template <class K, class V, class U>
V* TValHashMapT<K,V,U>::Find (const K& k)
{
  TValHashMapEntryT<K,V> *entry;
  unsigned int hash = U::Hash (k);
  
  entry = myBuckets[hash & myBucketMask];
  while (entry)
    {
      if (entry->hash == hash &&
	  U::Equal (k, entry->key))
	return &entry->value;
      
      entry = entry->next;
    }
  
  return 0;
}


template <class K, class V, class U>
void TValHashMapT<K,V,U>::Grow ()
{
  ReHash (myNumBuckets << 2);
  myShrinkSize = (myNumBuckets * BUCKET_MULTIPLIER) >> 4;
  myGrowSize <<= 2;
}


template <class K, class V, class U>
void TValHashMapT<K,V,U>::ReHash (int new_num_buckets)
{
  TValHashMapEntryT<K,V> **orig_buckets = myBuckets;
  TValHashMapEntryT<K,V> *entry, *old_entry, *tmp_entry;
  int old_num_buckets = myNumBuckets, index;
  
  myBuckets = (TValHashMapEntryT<K,V> **)
    calloc (new_num_buckets, sizeof (myBuckets));
  
  if (! myBuckets)
    throw TMemoryException ();
  
  myNumBuckets = new_num_buckets;
  myBucketMask = new_num_buckets - 1;
  
  for (int i = 0; i < old_num_buckets; i++)
    {
      entry = orig_buckets[i];
      while (entry)
	{
	  tmp_entry = entry;
	  entry = entry->next;
	  index = tmp_entry->hash & myBucketMask;
	  old_entry = myBuckets[index];
	  myBuckets[index] = tmp_entry;
	  tmp_entry->next = old_entry;	  
	}
    }
  
  free (orig_buckets);
}


template <class K, class V, class U>
void TValHashMapT<K,V,U>::Shrink ()
{
  ReHash (myNumBuckets >> 2);
  myShrinkSize >>= 2;
  myGrowSize >>= 2;
}

#endif // _T_VAL_HASH_MAP_T_HH
